import greenfoot.Color;
import greenfoot.Greenfoot;
import greenfoot.UserInfo;
import packages.event.Action;
import packages.geometry.Vector2D;
import packages.ui.Button;
import packages.ui.Text;
import packages.ui.UIWorld;

public class Game extends UIWorld {

    private static int lastDifficulty = 3;

    final Statistics stats = new Statistics();
    private final Button[][] buttons;
    private final Text scoreText;

    private Vector2D currentButton;
    private long lastClick;

    public Game() {
        this(lastDifficulty);
    }
    public Game(int difficulty) {
        super(600, 400, 1);
        stats.difficulty = difficulty;
        lastDifficulty = difficulty;
        getBackground().setColor(MainMenu.BKG_COLOR);
        getBackground().fill();

        buttons = new Button[10][difficulty];
        for(int i=0; i<10; i++) for(int j=0; j<difficulty; j++) {
            buttons[i][j] = new Button(new Text("", Color.BLACK, Color.LIGHT_GRAY), 30, 30);
            final int x = i, y = j;
            buttons[i][j].addClickAction(info -> buttonClicked(x, y));
        }
        scoreText = new Text("Score: 0", 25, Color.GRAY);

        //addFps();
        add(new Countdown(info -> startGame()), 0.5, 0.5);
    }



    private void startGame() {
        for(int i=0; i<10; i++) for(int j=0; j<stats.difficulty; j++) {
            int x = getWidth() / 2 + (int)((i - 5 + 0.5) * 35);
            int y = getHeight() / 2 + (int)((j - stats.difficulty * 0.5 + 0.5) * 35);
            addObject(buttons[i][j], x, y);
        }

        add(scoreText, 0.5, 0.8 + 0.0165 * stats.difficulty);

        Button b = new Button("Main menu");
        b.addClickAction(info -> Greenfoot.setWorld(new MainMenu()));
        add(b, 0.08, 0.03);

        b = new Button("Restart", b.getImage().getWidth(), b.getImage().getHeight());
        b.addClickAction(info -> restart());
        add(b, 0.92, 0.03);

        add(new Timer(30, info -> endGame()), 0.5, 0.2 - 0.0165 * stats.difficulty);

        stats.score = 0;
        colorDifferentButton();
    }



    private void buttonClicked(int x, int y) {
        lastClick = System.nanoTime();

        if(currentButton == null) {
            colorDifferentButton();
            return;
        }
        stats.totalClicks++;
        if(x == currentButton.x && y == currentButton.y) stats.score++;
        else {
            if(x != -1) stats.wrongClicks++;
            else stats.missedClicks++;
            if(stats.score > 0) stats.score--;
        }
        scoreText.setContent("Score: " + stats.score);

        colorDifferentButton();
    }

    private void colorDifferentButton() {
        if(currentButton != null)  buttons[(int)currentButton.x][(int)currentButton.y].getText().setBackgroundColor(Color.LIGHT_GRAY);
        currentButton = new Vector2D((int)(Math.random() * 10), (int)(Math.random() * stats.difficulty));
        buttons[(int)currentButton.x][(int)currentButton.y].getText().setBackgroundColor(Color.CYAN);
    }



    private void restart() {
        Greenfoot.setWorld(new Game(stats.difficulty));
    }

    private void endGame() {
        try{
            UserInfo user = UserInfo.getMyInfo();
            if(stats.realScore() > user.getScore()) {
                user.setScore(stats.realScore());
                user.setInt(0, stats.difficulty);
                user.store();
            }
        }
        catch(Exception e) {
            System.out.println("Score could not be saved");
        }
        Greenfoot.setWorld(new Result(stats));
    }


    @Override
    public void act() {
        if(getObjects(Countdown.class).size() > 0) return;
        if(System.nanoTime() > lastClick + maxTime()) buttonClicked(-1, -1);
    }


    private long maxTime() {
        return 3000000000l - stats.difficulty * 200000000;
    }










    private static class Countdown extends Text {
        private long start;
        private final Action action;
        public Countdown(Action action) {
            this(3, action);
        }
        public Countdown(int seconds, Action action) {
            super(seconds + "", 40, Color.RED);
            this.action = action;
            start = System.nanoTime() + seconds * 1000000000l;
        }


        @Override
        public void act() {
            if(getContent().equals("START!")) {
                if(System.nanoTime() > start + 500000000) {
                    getWorld().removeObject(this);
                    action.run(null);
                    return;
                }
            }
            else if(System.nanoTime() >= start) {
                setFontSize(40);
                setContent("START!");
            }
            else {
                int remaining = (int)Math.ceil((start - System.nanoTime()) / 1000000000d);
                setContent(remaining + "");
                int fontSize = 20 + (int)(20 * (((start - System.nanoTime()) / 1000000000d) % 1));
                setFontSize(fontSize);
            }
        }
    }
    
    private static class Timer extends Text {
        private long start;
        private final Action action;
        public Timer(int seconds, Action action) {
            super(seconds + "", 40, Color.LIGHT_GRAY);
            this.action = action;
            start = System.nanoTime() + seconds * 1000000000l;
        }


        @Override
        public void act() {
            if(System.nanoTime() >= start) {
                getWorld().removeObject(this);
                action.run(null);
                return;
            }
            else {
                int remaining = (int)Math.ceil((start - System.nanoTime()) / 1000000000d);
                if(remaining <= 3) setColor(Color.RED);
                setContent(remaining + "");
            }
        }
    }


    public static class Statistics {
        public int totalClicks, wrongClicks, missedClicks, difficulty, score;

        public int realScore() {
            return difficulty * score;
        }
    }
}